import os

import numpy as np
import torch.utils.data


class SearchDataset(torch.utils.data.Dataset):
    def __init__(self, transform=None):
        self.transform = transform

    def __len__(self):
        return int(os.environ.get("AUTOALBUMENT_TEST_DATASET_LENGTH", 16))

    def __getitem__(self, index):
        np.random.seed(index)
        image = np.random.randint(low=0, high=256, size=(32, 32, 3), dtype=np.uint8)
        label = np.random.randint(low=0, high=10, dtype=np.int64)

        if self.transform is not None:
            transformed = self.transform(image=image)
            image = transformed["image"]

        return image, label



import random

import torch
import numpy as np


def np_to_pytorch(np_array):
    return torch.from_numpy(np_array.transpose(0, 3, 1, 2))


def torch_to_np(torch_tensor):
    return torch_tensor.numpy().transpose(0, 2, 3, 1)


def assert_batches_match(np_images, pytorch_batch, rtol=1.0e-4, atol=1.0e-8):
    np_batch = torch_to_np(pytorch_batch)
    for np_image, batch_image in zip(np_images, np_batch):
        assert np.allclose(np_image, batch_image, rtol=rtol, atol=atol)



import numpy as np
import pytest
import torch

from tests.utils import np_to_pytorch


@pytest.fixture
def image_batches(batch_size=4):
    np_images = np.random.uniform(low=0.0, high=1.0, size=(batch_size, 128, 128, 3)).astype("float32")
    pytorch_batch = np_to_pytorch(np_images)
    return np_images, pytorch_batch


@pytest.fixture
def gradcheck_batch(batch_size=4, num_channels=3, height=4, width=4):
    return torch.rand(batch_size, num_channels, height, width).type(torch.float64)



import torch


def get_combined_parameter_values(state_dict, prefixes):
    combined_parameter_values = []
    for prefix in prefixes:
        weights = state_dict[f"{prefix}._weights"]
        temperature = torch.zeros(weights.size(), dtype=weights.dtype).to(weights.device)
        for i in range(len(weights)):
            temperature_key = f"{prefix}.operations.{i}.temperature"
            temperature[i] = state_dict[temperature_key]
        weights = torch.div(weights, temperature).softmax(0)
        for i, weight in enumerate(weights):
            probability_key = f"{prefix}.operations.{i}._probability"
            magnitude_key = f"{prefix}.operations.{i}._magnitude"
            probability = (weight * state_dict[probability_key].clamp(0.0, 1.0)).item()
            magnitude = state_dict.get(magnitude_key)
            if magnitude is not None:
                magnitude = magnitude.clamp(0.0, 1.0).item()
            value = probability * magnitude if magnitude is not None else probability
            combined_parameter_values.append(value)
    return combined_parameter_values


def get_average_parameter_change(state_dict_1, state_dict_2):
    prefixes = {".".join(key.split(".")[:4]) for key in state_dict_1.keys() if key.startswith("sub_policies")}
    values_1 = get_combined_parameter_values(state_dict_1, prefixes)
    values_2 = get_combined_parameter_values(state_dict_2, prefixes)
    values_change = [abs(v2 - v1) for v1, v2 in zip(values_1, values_2)]
    return sum(values_change) / len(values_change)



from typing import Tuple

import segmentation_models_pytorch as smp
import timm
from torch import nn, Tensor
from torch.nn import Flatten


class BaseDiscriminator(nn.Module):
    def __init__(self, *args, **kwargs):
        super().__init__()

    def forward(self, input: Tensor) -> Tuple[Tensor, Tensor]:
        raise NotImplementedError


class ClassificationModel(BaseDiscriminator):
    def __init__(self, architecture, pretrained, num_classes):
        super().__init__()
        self.base_model = timm.create_model(architecture, pretrained=pretrained)
        self.base_model.reset_classifier(num_classes)
        self.classifier = self.base_model.get_classifier()
        num_features = self.classifier.in_features
        self.discriminator = nn.Sequential(
            nn.Linear(num_features, num_features), nn.ReLU(), nn.Linear(num_features, 1)
        )

    def forward(self, input: Tensor) -> Tuple[Tensor, Tensor]:
        x = self.base_model.forward_features(input)
        x = self.base_model.global_pool(x).flatten(1)
        return self.classifier(x), self.discriminator(x).view(-1)


class SemanticSegmentationModel(BaseDiscriminator):
    def __init__(self, architecture, encoder_architecture, num_classes, pretrained):
        super().__init__()
        model = getattr(smp, architecture)

        self.base_model = model(
            encoder_architecture, encoder_weights=self._get_encoder_weights(pretrained), classes=num_classes
        )
        num_features = self.base_model.encoder.out_channels[-1]
        self.base_model.classification_head = nn.Sequential(
            nn.AdaptiveAvgPool2d(1),
            Flatten(),
            nn.Linear(num_features, num_features),
            nn.ReLU(),
            nn.Linear(num_features, 1),
        )

    @staticmethod
    def _get_encoder_weights(pretrained):
        if isinstance(pretrained, bool):
            return "imagenet" if pretrained else None
        return pretrained

    def forward(self, input: Tensor) -> Tuple[Tensor, Tensor]:
        mask, discriminator_output = self.base_model(input)
        return mask, discriminator_output.view(-1)



import copy
import os
import sys

import hydra
from hydra.core.config_store import ConfigStore
from omegaconf import OmegaConf

from autoalbument.config.faster_autoaugment import FasterAutoAugmentSearchConfig
from autoalbument.config.validation import validate_cfg
from autoalbument.faster_autoaugment.search import get_faa_searcher
from autoalbument.utils.hydra import get_config_dir

OmegaConf.register_resolver("config_dir", get_config_dir)
cs = ConfigStore.instance()
cs.store(name="config", node=FasterAutoAugmentSearchConfig)


def get_prettified_cfg(cfg):
    cfg = copy.deepcopy(cfg)
    cfg = OmegaConf.to_container(cfg, resolve=True)

    if cfg["task"] == "classification":
        del cfg["semantic_segmentation_model"]
    else:
        del cfg["classification_model"]

    return OmegaConf.to_yaml(cfg)


@hydra.main(config_path="conf", config_name="config")
def main(cfg):
    validate_cfg(cfg)
    config_dir = get_config_dir()
    if config_dir is not None:
        sys.path.append(config_dir)
    print(get_prettified_cfg(cfg))
    cwd = os.getcwd()
    print(f"Working directory: {cwd}")
    faa_searcher = get_faa_searcher(cfg)
    faa_searcher.search()



from dataclasses import dataclass, field
from typing import List, Optional, Any, Dict

from omegaconf import MISSING


@dataclass
class NormalizationConfig:
    mean: List[float] = field(default_factory=lambda: [0.485, 0.456, 0.406])
    std: List[float] = field(default_factory=lambda: [0.229, 0.224, 0.225])


@dataclass
class DataloaderConfig:
    batch_size: int = 128
    shuffle: bool = True
    num_workers: int = 4


@dataclass
class DataConfig:
    normalization: NormalizationConfig = NormalizationConfig()
    dataloader: Any = MISSING
    preprocessing: Optional[Any] = None
    dataset_file: Optional[Any] = None
    dataset: Optional[Any] = None
    input_dtype: str = "uint8"


@dataclass
class OptimConfig:
    epochs: int = 20
    main: Any = MISSING
    policy: Any = MISSING


@dataclass
class PolicyModelConfig:
    task_factor: float = 0.1
    gp_factor: float = 10
    temperature: float = 0.05
    num_sub_policies: int = 150
    num_chunks: int = 8
    operation_count: int = 4


@dataclass
class ClassificationModelConfig:
    _target_: str = "autoalbument.faster_autoaugment.models.ClassificationModel"
    num_classes: Optional[int] = None
    architecture: str = "resnet18"
    pretrained: bool = False


@dataclass
class SemanticSegmentationModelConfig:
    _target_: str = "autoalbument.faster_autoaugment.models.SemanticSegmentationModel"
    num_classes: Optional[int] = None
    architecture: str = "Unet"
    encoder_architecture: str = "resnet18"
    pretrained: bool = False


@dataclass
class FasterAutoAugmentSearchConfig:
    policy_model: PolicyModelConfig = PolicyModelConfig()
    optim: OptimConfig = OptimConfig(main=MISSING, policy=MISSING)
    device: str = "cuda"
    task: str = MISSING
    cudnn_benchmark: bool = True
    save_checkpoints: bool = False
    checkpoint_path: Optional[str] = None
    tensorboard_logs_dir: Optional[str] = None
    classification_model: Optional[ClassificationModelConfig] = None
    semantic_segmentation_model: Optional[SemanticSegmentationModelConfig] = None
    data: DataConfig = DataConfig(dataloader=MISSING)
    seed: Optional[int] = None



import argparse
import os

from hydra.utils import to_absolute_path


def get_config_dir():
    parser = argparse.ArgumentParser()
    parser.add_argument("--config-dir", required=False)
    args, unknown = parser.parse_known_args()
    config_dir = args.config_dir
    if config_dir is None:
        config_dir = os.environ.get("AUTOALBUMENT_CONFIG_DIR")
    return to_absolute_path(config_dir)


def get_dataset_filepath(dataset_file):
    if os.path.isabs(dataset_file):
        return dataset_file

    base_path = get_config_dir()
    if base_path is None:
        raise ValueError(
            f"Couldn't deduce the value for `config_dir` (a directory that contains the search.yaml configuration "
            f"file) to locate the {dataset_file} dataset file. Please use one of the following alternatives to fix "
            f"the error: "
            f"\n- Run `autoalbument-search` with the `--config-dir` parameter."
            f"\n- Use the absolute filepath for the `data.dataset_file` configuration parameter."
            f"\n- Provide the value for `config_dir` through the `AUTOALBUMENT_CONFIG_DIR` environment variable, e.g. "
            f"`export AUTOALBUMENT_CONFIG_DIR=/path/to/config/dir"
        )

    return os.path.join(base_path, dataset_file)



"""
Based on the official implementation of Faster AutoAugment
https://github.com/moskomule/dda/blob/master/faster_autoaugment/search.py
"""

import copy
import logging
import os
from pathlib import Path
from typing import Tuple

import albumentations as A
import torch
from albumentations.pytorch import ToTensorV2
from hydra.utils import instantiate
from torch import Tensor, nn
from torch.utils.tensorboard import SummaryWriter
from tqdm import tqdm

from autoalbument.faster_autoaugment.metrics import get_average_parameter_change
from autoalbument.faster_autoaugment.models import ClassificationModel, SemanticSegmentationModel
from autoalbument.faster_autoaugment.utils import MAX_VALUES_BY_INPUT_DTYPE, get_dataset_cls, MetricTracker, set_seed
from autoalbument.utils.files import symlink
from autoalbument.faster_autoaugment.policy import Policy

log = logging.getLogger(__name__)


class FasterAutoAugmentBase:
    def __init__(self, cfg):
        self.cfg = cfg
        torch.backends.cudnn.benchmark = self.cfg.cudnn_benchmark
        self.start_epoch = 1
        self.set_seed()
        self.dataloader = self.create_dataloader()
        self.models = self.create_models()
        self.optimizers = self.create_optimizers()
        self.loss = self.create_loss()
        self.metric_tracker = self.create_metric_tracker()
        self.paths = self.create_directories()
        self.tensorboard_writer = self.create_tensorboard_writer()
        self.epoch = None
        if self.cfg.checkpoint_path:
            self.load_checkpoint()
        self.saved_policy_state_dict = self.get_policy_state_dict()

    def set_seed(self):
        seed = getattr(self.cfg, "seed", None)
        if seed is not None:
            set_seed(seed)

    def create_tensorboard_writer(self):
        if self.cfg.tensorboard_logs_dir:
            return SummaryWriter(os.path.join(self.cfg.tensorboard_logs_dir, os.getcwd().replace(os.sep, ".")))
        return None

    def get_policy_state_dict(self):
        state_dict = copy.deepcopy(self.models["policy"].state_dict())
        for k, v in state_dict.items():
            state_dict[k] = v.to("cpu")
        return state_dict

    def create_directories(self):
        policy_dir = Path.cwd() / "policy"
        policy_dir.mkdir(exist_ok=True)
        if self.cfg.save_checkpoints:
            checkpoints_dir = Path.cwd() / "checkpoints"
            checkpoints_dir.mkdir(exist_ok=True)
        else:
            checkpoints_dir = None
        return {
            "policy_dir": policy_dir,
            "checkpoints_dir": checkpoints_dir,
            "latest_policy_path": policy_dir / "latest.json",
            "latest_checkpoint_path": checkpoints_dir / "latest.pth" if checkpoints_dir is not None else None,
        }

    def get_latest_policy_save_path(self):
        return

    def create_metric_tracker(self):
        return MetricTracker()

    def get_preprocessing_transforms(self):
        preprocessing_config = self.cfg.data.preprocessing
        preprocessing_transforms = []
        if preprocessing_config:
            for preprocessing_transform in preprocessing_config:
                for transform_name, transform_args in preprocessing_transform.items():
                    transform = A.from_dict(
                        {
                            "transform": {
                                "__class_fullname__": "albumentations.augmentations.transforms." + transform_name,
                                **transform_args,
                            }
                        }
                    )
                    preprocessing_transforms.append(transform)
        return preprocessing_transforms

    def create_preprocessing_transform(self):
        normalization_config = self.cfg.data.normalization
        input_dtype = self.cfg.data.input_dtype
        transform = A.Compose(
            [
                *self.get_preprocessing_transforms(),
                A.Normalize(
                    mean=normalization_config.mean,
                    std=normalization_config.std,
                    max_pixel_value=MAX_VALUES_BY_INPUT_DTYPE[input_dtype],
                ),
                ToTensorV2(transpose_mask=True),
            ]
        )
        log.info(f"Preprocessing transform:\n{transform}")
        return transform

    def create_dataloader(self):
        transform = self.create_preprocessing_transform()

        data_config = self.cfg.data

        if getattr(data_config, "dataset", None):
            dataset = instantiate(data_config.dataset, transform=transform)
        elif getattr(data_config, "dataset_file", None):
            dataset_cls = get_dataset_cls(self.cfg.data.dataset_file)
            dataset = dataset_cls(transform=transform)
        else:
            raise ValueError(f"You should provide a correct dataset in data.dataset, got {data_config.dataset}")

        dataloader = instantiate(self.cfg.data.dataloader, dataset=dataset)
        return dataloader

    def create_optimizers(self):
        optimizer_config = self.cfg.optim
        main_optimizer = instantiate(optimizer_config.main, params=self.models["main"].parameters())
        policy_optimizer = instantiate(optimizer_config.policy, params=self.models["policy"].parameters())
        return {
            "main": main_optimizer,
            "policy": policy_optimizer,
        }

    def get_main_model_cfg(self):
        raise NotImplementedError

    def create_main_model(self):
        model_cfg = self.get_main_model_cfg()
        main_model = instantiate(model_cfg)
        main_model = main_model.to(self.cfg.device)
        return main_model

    def create_policy_model(self):
        policy_model_cfg = self.cfg.policy_model
        normalization_cfg = self.cfg.data.normalization
        policy_model = Policy.faster_auto_augment_policy(
            policy_model_cfg.num_sub_policies,
            policy_model_cfg.temperature,
            policy_model_cfg.operation_count,
            policy_model_cfg.num_chunks,
            mean=torch.tensor(normalization_cfg.mean),
            std=torch.tensor(normalization_cfg.std),
        ).to(self.cfg.device)
        return policy_model

    def create_models(self):
        main_model = self.create_main_model()
        policy_model = self.create_policy_model()
        return {"main": main_model, "policy": policy_model}

    def policy_forward_for_policy_train(self, a_input, a_target):
        raise NotImplementedError

    def gradient_penalty(self, real: Tensor, fake: Tensor) -> Tensor:
        alpha = real.new_empty(real.size(0), 1, 1, 1).uniform_(0, 1)
        interpolated = alpha * real + (1 - alpha) * fake
        interpolated.requires_grad_()
        _, output = self.models["main"](interpolated)
        grad = torch.autograd.grad(
            outputs=output,
            inputs=interpolated,
            grad_outputs=torch.ones_like(output),
            create_graph=True,
            retain_graph=True,
            only_inputs=True,
        )[0]
        return (grad.norm(2, dim=1) - 1).pow(2).mean()

    def create_loss(self):
        return nn.CrossEntropyLoss().to(self.cfg.device)

    def wgan_loss(
        self, n_input: Tensor, n_target: Tensor, a_input: Tensor, a_target: Tensor
    ) -> Tuple[Tensor, Tensor, Tensor]:
        ones = n_input.new_tensor(1.0)
        self.models["main"].requires_grad_(True)
        self.models["main"].zero_grad()
        output, n_output = self.models["main"](n_input)
        loss = self.cfg.policy_model.task_factor * self.loss(output, n_target)
        loss.backward(retain_graph=True)
        d_n_loss = n_output.mean()

        d_n_loss.backward(-ones)

        with torch.no_grad():
            a_input = self.models["policy"].denormalize_(a_input)
            augmented = self.models["policy"]({"image_batch": a_input})["image_batch"]

        _, a_output = self.models["main"](augmented)
        d_a_loss = a_output.mean()

        d_a_loss.backward(ones)
        gp = self.cfg.policy_model.gp_factor * self.gradient_penalty(n_input, augmented)
        gp.backward()
        self.optimizers["main"].step()

        self.models["main"].requires_grad_(False)
        self.models["policy"].zero_grad()

        augmented_input, maybe_augmented_target = self.policy_forward_for_policy_train(a_input, a_target)

        _output, a_output = self.models["main"](augmented_input)
        _loss = self.cfg.policy_model.task_factor * self.loss(_output, maybe_augmented_target)
        _loss.backward(retain_graph=True)
        a_loss = a_output.mean()
        a_loss.backward(-ones)
        self.optimizers["policy"].step()
        return loss + _loss, -d_n_loss + d_a_loss + gp, -a_loss

    def train_step(self, input, target):
        b = input.size(0) // 2
        a_input, a_target = input[:b], target[:b]
        n_input, n_target = input[b:], target[b:]
        loss, d_loss, a_loss = self.wgan_loss(n_input, n_target, a_input, a_target)
        return {
            "loss": loss,
            "d_loss": d_loss,
            "a_loss": a_loss,
        }

    def get_progress_bar_description(self):
        return f"Epoch: {self.epoch}. {self.metric_tracker}"

    def train_epoch(self):
        self.metric_tracker.reset()
        pbar = tqdm(total=len(self.dataloader))
        for input, target in self.dataloader:
            input = input.to(self.cfg.device)
            target = target.to(self.cfg.device)
            loss_dict = self.train_step(input, target)
            self.metric_tracker.update(loss_dict)
            pbar.set_description(self.get_progress_bar_description())
            pbar.update()

        average_parameter_change = get_average_parameter_change(
            self.saved_policy_state_dict, self.get_policy_state_dict()
        )
        self.saved_policy_state_dict = self.get_policy_state_dict()
        self.metric_tracker.update({"Average Parameter change": average_parameter_change})
        pbar.set_description(self.get_progress_bar_description())
        pbar.close()

    def write_tensorboard_logs(self):
        if not self.tensorboard_writer:
            return
        for metric, avg_value in self.metric_tracker.get_avg_values():
            self.tensorboard_writer.add_scalar(metric, avg_value, self.epoch)

    def save_policy(self):

        transform = self.models["policy"].create_transform(
            input_dtype=self.cfg.data.input_dtype,
            preprocessing_transforms=self.get_preprocessing_transforms(),
        )
        policy_save_path = self.paths["policy_dir"] / f"epoch_{self.epoch}.json"
        A.save(transform, str(policy_save_path))
        symlink(policy_save_path, self.paths["latest_policy_path"])
        log.info(
            f"Policy is saved to {policy_save_path}. "
            f"{self.paths['latest_policy_path']} now also points to this policy file."
        )

    def save_checkpoint(self):
        checkpoint = {
            "epoch": self.epoch,
            "models": {k: v.state_dict() for k, v in self.models.items()},
            "optimizers": {k: v.state_dict() for k, v in self.optimizers.items()},
        }
        checkpoint_save_path = self.paths["checkpoints_dir"] / f"epoch_{self.epoch}.pth"
        torch.save(checkpoint, checkpoint_save_path)
        symlink(checkpoint_save_path, self.paths["latest_checkpoint_path"])
        log.info(
            f"Checkpoint is saved to {checkpoint_save_path}. "
            f"{self.paths['latest_checkpoint_path']} now also points to this checkpoint file."
        )

    def load_checkpoint(self):
        checkpoint_path = self.cfg.checkpoint_path
        log.info(f"Loading checkpoint {checkpoint_path}")
        checkpoint = torch.load(checkpoint_path, map_location=self.cfg.device)
        for name, model in self.models.items():
            model.load_state_dict(checkpoint["models"][name])
        for name, optimizer in self.optimizers.items():
            optimizer.load_state_dict(checkpoint["optimizers"][name])
        self.start_epoch = checkpoint["epoch"] + 1
        self.saved_policy_state_dict = self.get_policy_state_dict()

    def save(self):
        self.save_policy()
        if self.cfg.save_checkpoints:
            self.save_checkpoint()

    def get_search_summary(self):
        summary = [
            f"\n\nSearch is finished.\n"
            f"- Configuration files for policies found at each epoch are located in {self.paths['policy_dir']}."
        ]
        if self.cfg.save_checkpoints:
            summary.append(
                f"- Checkpoint files for models and optimizers at each epoch are located in "
                f"{self.paths['checkpoints_dir']}."
            )

        latest_policy_path = self.paths["latest_policy_path"]
        load_command = f'transform = A.load("{latest_policy_path}")\n'

        separator = "-" * len(load_command)
        if self.cfg.task == "classification":
            transform_text = "transformed = transform(image=image)\n" 'transformed_image = transformed["image"]\n'
        else:
            transform_text = (
                "transformed = transform(image=image, mask=mask)\n"
                'transformed_image = transformed["image"]\n'
                'transformed_mask = transformed["mask"]\n'
            )

        summary.append(
            f"\nUse Albumentations to load the found policies and transform images:\n"
            f"{separator}\n"
            f"import albumentations as A\n\n"
            f'transform = A.load("{latest_policy_path}")\n' + transform_text + f"{separator}\n"
        )

        return "\n".join(summary)

    def search(self):
        for epoch in range(self.start_epoch, self.cfg.optim.epochs + 1):
            self.epoch = epoch
            self.train_epoch()
            self.write_tensorboard_logs()
            self.save()

        log.info(self.get_search_summary())


class FAAClassification(FasterAutoAugmentBase):
    def get_main_model_cfg(self):
        return self.cfg.classification_model

    def create_loss(self):
        return nn.CrossEntropyLoss().to(self.cfg.device)

    def policy_forward_for_policy_train(self, a_input, a_target):
        output = self.models["policy"]({"image_batch": a_input})["image_batch"]
        return output, a_target


class FAASemanticSegmentation(FasterAutoAugmentBase):
    def get_main_model_cfg(self):
        return self.cfg.semantic_segmentation_model

    def create_loss(self):
        return nn.BCEWithLogitsLoss().to(self.cfg.device)

    def policy_forward_for_policy_train(self, a_input, a_target):
        output = self.models["policy"]({"image_batch": a_input, "mask_batch": a_target})
        return output["image_batch"], output["mask_batch"]


def get_faa_searcher(cfg):
    task = cfg.task
    if task == "semantic_segmentation":
        return FAASemanticSegmentation(cfg)
    elif task == "classification":
        return FAAClassification(cfg)
    raise ValueError(f"Unsupported task: {task}. Supported tasks: classification, semantic_segmentation.")



def validate_cfg(cfg):
    data_cfg = cfg.data
    if (data_cfg.dataset is None) == (data_cfg.dataset_file is None):
        raise ValueError(
            f"You should provide either data.dataset or data.dataset_file but not both. "
            f"Got {data_cfg.dataset} for data.dataset and {data_cfg.dataset_file} for data.dataset_file."
        )



import torch
import torch.nn.functional as F


def get_scaling_matrix(img_batch, dx=0, dy=0):
    height, width = img_batch.shape[-2:]
    matrix = torch.eye(2, 3, dtype=img_batch.dtype, device=img_batch.device)
    matrix[0, 2] = dx * width
    matrix[1, 2] = dy * height
    return matrix


def get_normalization_matrix(img_batch):
    height, width = img_batch.shape[-2:]
    matrix = torch.zeros(3, 3, dtype=img_batch.dtype, device=img_batch.device)
    matrix[0, 0] = 2.0 / width
    matrix[0, 1] = 0
    matrix[1, 1] = 2.0 / height
    matrix[1, 0] = 0
    matrix[0, -1] = -1.0
    matrix[1, -1] = -1.0
    matrix[-1, -1] = 1.0
    return matrix


def get_rotation_matrix(img_batch, angle=None, scale=None):
    matrix = torch.zeros(2, 3, dtype=img_batch.dtype, device=img_batch.device)
    height, width = img_batch.shape[-2:]
    if angle is None:
        angle = torch.tensor([0.0], dtype=img_batch.dtype, device=img_batch.device)
    if scale is None:
        scale = torch.tensor([1.0], dtype=img_batch.dtype, device=img_batch.device)

    center_x = width / 2
    center_y = height / 2
    angle_in_radians = torch.deg2rad(angle)
    alpha = torch.cos(angle_in_radians) * scale
    beta = torch.sin(angle_in_radians) * scale
    matrix[0, 0] = alpha
    matrix[0, 1] = beta
    matrix[0, 2] = (1.0 - alpha) * center_x - beta * center_y
    matrix[1, 0] = -beta
    matrix[1, 1] = alpha
    matrix[1, 2] = beta * center_x + (1.0 - alpha) * center_y
    return matrix


def convert_2x3_affine_matrix_to_3x3(matrix):
    pad = torch.tensor([[0, 0, 1]], dtype=matrix.dtype, device=matrix.device)
    return torch.cat([matrix, pad])


def expand_matrix_at_batch_dimension(matrix, batch_size):
    return matrix.expand(batch_size, -1, -1)


def warp_affine(img_batch, affine_matrix, padding_mode, align_corners=True):
    matrix = convert_2x3_affine_matrix_to_3x3(affine_matrix)
    normalization_matrix = get_normalization_matrix(img_batch)

    batch_size = img_batch.shape[0]
    matrix = expand_matrix_at_batch_dimension(matrix, batch_size)
    normalization_matrix = expand_matrix_at_batch_dimension(normalization_matrix, batch_size)

    inverse_normalization_matrix = torch.inverse(normalization_matrix)
    theta_3x3 = torch.inverse(normalization_matrix @ matrix @ inverse_normalization_matrix)
    theta = theta_3x3[:, :2, :]

    grid = F.affine_grid(theta, img_batch.shape, align_corners=align_corners)
    return F.grid_sample(img_batch, grid, padding_mode=padding_mode, align_corners=align_corners)



import albumentations.augmentations.functional as F
import pytest
import torch
from torch.autograd import gradcheck

import autoalbument.faster_autoaugment.albumentations_pytorch.functional as PF
from tests.utils import assert_batches_match


class Base:
    def scalar_to_tensor(self, arg, requires_grad=False, dtype=torch.float32):
        if arg is None:
            return None
        return torch.tensor(arg, requires_grad=requires_grad, dtype=dtype)

    def test_albumentations_match(self, image_batches, arg):
        np_images, pytorch_batch = image_batches
        tensor_arg = self.scalar_to_tensor(arg)
        augmented_np_images = [self.albumentations_fn(image, arg) for image in np_images]
        augmented_pytorch_batch = self.albumentations_pytorch_fn(pytorch_batch, tensor_arg)
        assert_batches_match(augmented_np_images, augmented_pytorch_batch)

    def test_gradients(self, gradcheck_batch, arg):
        tensor_arg = self.scalar_to_tensor(arg, requires_grad=True, dtype=torch.float64)
        gradcheck(self.albumentations_pytorch_fn, (gradcheck_batch, tensor_arg))

    def albumentations_fn(self, image, arg):
        raise NotImplementedError

    def albumentations_pytorch_fn(self, pytorch_batch, arg):
        raise NotImplementedError


@pytest.mark.parametrize("arg", [0.2, 0.4, 0.8])
class TestSolarize(Base):
    def albumentations_fn(self, image, arg):
        return F.solarize(image, threshold=arg)

    def albumentations_pytorch_fn(self, pytorch_batch, arg):
        return PF.solarize(pytorch_batch, threshold=arg)

    def test_gradients(self, gradcheck_batch, arg):
        pass


@pytest.mark.parametrize("arg", [[0.0, 0.0, 0.0], [1.0, 1.0, 1.0], [-1.0, -1.0, -1.0], [0.0, 0.7, -0.2]])
class TestShiftRgb(Base):
    def albumentations_fn(self, image, arg):
        return F.shift_rgb(image, r_shift=arg[0], g_shift=arg[1], b_shift=arg[2])

    def albumentations_pytorch_fn(self, pytorch_batch, arg):
        return PF.shift_rgb(pytorch_batch, r_shift=arg[0], g_shift=arg[1], b_shift=arg[2])


@pytest.mark.parametrize("arg", [-1.0, 0.1, 0.5, 1.0])
class TestBrightnessAdjust(Base):
    def albumentations_fn(self, image, arg):
        return F.brightness_contrast_adjust(image, beta=arg, beta_by_max=True)

    def albumentations_pytorch_fn(self, pytorch_batch, arg):
        return PF.brightness_adjust(pytorch_batch, beta=arg)


@pytest.mark.parametrize("arg", [-1.0, 0.1, 0.5, 1.0])
class TestContrastAdjust(Base):
    def albumentations_fn(self, image, arg):
        return F.brightness_contrast_adjust(image, alpha=arg)

    def albumentations_pytorch_fn(self, pytorch_batch, arg):
        return PF.contrast_adjust(pytorch_batch, alpha=arg)


@pytest.mark.parametrize("arg", [None])
class TestVflip(Base):
    def albumentations_fn(self, image, arg):
        return F.vflip(image)

    def albumentations_pytorch_fn(self, pytorch_batch, arg):
        return PF.vflip(pytorch_batch)

    def test_gradients(self, gradcheck_batch, arg):
        pass


@pytest.mark.parametrize("arg", [None])
class TestHflip(Base):
    def albumentations_fn(self, image, arg):
        return F.hflip(image)

    def albumentations_pytorch_fn(self, pytorch_batch, arg):
        return PF.hflip(pytorch_batch)

    def test_gradients(self, gradcheck_batch, arg):
        pass


@pytest.mark.parametrize(
    "arg",
    [
        [0.01],
        [-0.5],
        [0.5],
        [1.0 - 1e-6],
        [-1.0 + 1e-6],
    ],
)
class TestShiftX(Base):
    def albumentations_pytorch_fn(self, pytorch_batch, arg):
        return PF.shift_x(pytorch_batch, dx=arg)

    def test_albumentations_match(self, image_batches, arg):
        pass


@pytest.mark.parametrize(
    "arg",
    [
        [0.01],
        [-0.5],
        [0.5],
        [1.0 - 1e-6],
        [-1.0 + 1e-6],
    ],
)
class TestShiftY(Base):
    def albumentations_pytorch_fn(self, pytorch_batch, arg):
        return PF.shift_y(pytorch_batch, dy=arg)

    def test_albumentations_match(self, image_batches, arg):
        pass


@pytest.mark.parametrize(
    "arg",
    [
        [0.1],
        [-0.5],
        [0.5],
        [1.0 - 1e-6],
        [-1.0 + 1e-6],
    ],
)
class TestScale(Base):
    def albumentations_pytorch_fn(self, pytorch_batch, arg):
        return PF.scale(pytorch_batch, scale=arg)

    def test_albumentations_match(self, image_batches, arg):
        pass


@pytest.mark.parametrize(
    "arg",
    [
        [0.1],
        [-0.5],
        [0.5],
        [1.0 - 1e-6],
        [-1.0 + 1e-6],
    ],
)
class TestRotate(Base):
    def albumentations_pytorch_fn(self, pytorch_batch, arg):
        return PF.rotate(pytorch_batch, angle=arg)

    def test_albumentations_match(self, image_batches, arg):
        pass



import shutil
import tempfile
import platform
import os


def symlink(src, dst):
    if platform.system == "Windows":
        shutil.copy2(src, dst)
    else:
        tmp_dst = tempfile.mktemp(dir=dst.parents[0])
        os.symlink(src, tmp_dst)
        os.replace(tmp_dst, dst)



import os
from pathlib import Path

from hydra.experimental import initialize, compose, initialize_config_dir

from autoalbument.faster_autoaugment.search import get_faa_searcher


def test_dataloader_drops_last(tmpdir) -> None:
    config_path = "./configs/classification"
    with initialize(config_path=str(config_path)):
        os.environ["AUTOALBUMENT_TEST_DATASET_LENGTH"] = "17"
        os.environ["AUTOALBUMENT_CONFIG_DIR"] = str((Path(__file__).parent / config_path).resolve())
        os.chdir(tmpdir)
        cfg = compose(config_name="search", overrides=["data.dataloader.batch_size=12"])
        faa_searcher = get_faa_searcher(cfg)
        faa_searcher.search()



import io
import os
import re

from setuptools import find_packages, setup


def get_version():
    current_dir = os.path.abspath(os.path.dirname(__file__))
    version_file = os.path.join(current_dir, "autoalbument", "__init__.py")
    with io.open(version_file, encoding="utf-8") as f:
        return re.search(r'^__version__ = [\'"]([^\'"]*)[\'"]', f.read(), re.M).group(1)


def get_long_description():
    base_dir = os.path.abspath(os.path.dirname(__file__))
    with io.open(os.path.join(base_dir, "README.md"), encoding="utf-8") as f:
        return f.read()


setup(
    name="autoalbument",
    version=get_version(),
    description="AutoML for image augmentation",
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    author="Alex Parinov, Vladimir Iglovikov, Eugene Khvedchenya, Druzhinin Mikhail, Buslaev Alexander",
    license="MIT",
    url="https://github.com/albumentations-team/autoalbument",
    install_requires=[
        "albumentations>=0.5.1",
        "torch>=1.6.0",
        "hydra-core>=1.0",
        "timm>=0.3.2",  # This version is required for segmentation-models-pytorch
        "segmentation-models-pytorch>=0.1.3",
        "tqdm",
        "click",
        "colorama",
        "tensorboard",
    ],
    entry_points={
        "console_scripts": [
            "autoalbument-create = autoalbument.cli.create:main",
            "autoalbument-search = autoalbument.cli.search:main",
        ],
    },
    extras_require={"tests": ["pytest"]},
    packages=find_packages(exclude=["tests"]),
    include_package_data=True,
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)
